package com.sun.xml.internal.ws.spi.db;

import java.util.NoSuchElementException;
import java.lang.reflect.Array;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.Collection;
import java.util.List;
import java.io.InputStream;
import javax.xml.transform.Source;
import javax.xml.bind.attachment.AttachmentUnmarshaller;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Result;
import org.xml.sax.ContentHandler;
import org.w3c.dom.Node;
import javax.xml.namespace.NamespaceContext;
import java.io.OutputStream;
import javax.xml.bind.JAXBException;
import javax.xml.bind.attachment.AttachmentMarshaller;
import javax.xml.stream.XMLStreamWriter;

public class RepeatedElementBridge<T> implements XMLBridge<T>
{
    XMLBridge<T> delegate;
    RepeatedElementBridge.CollectionHandler collectionHandler;
    static final RepeatedElementBridge.CollectionHandler ListHandler;
    static final RepeatedElementBridge.CollectionHandler HashSetHandler;
    
    public RepeatedElementBridge(TypeInfo typeInfo, XMLBridge xb) {
        this.delegate = xb;
        this.collectionHandler = create(typeInfo);
    }
    
    public RepeatedElementBridge.CollectionHandler collectionHandler() {
        return this.collectionHandler;
    }
    
    @Override
    public BindingContext context() {
        return this.delegate.context();
    }
    
    @Override
    public void marshal(T object, XMLStreamWriter output, AttachmentMarshaller am) throws JAXBException {
        this.delegate.marshal(object, output, am);
    }
    
    @Override
    public void marshal(T object, OutputStream output, NamespaceContext nsContext, AttachmentMarshaller am) throws JAXBException {
        this.delegate.marshal(object, output, nsContext, am);
    }
    
    @Override
    public void marshal(T object, Node output) throws JAXBException {
        this.delegate.marshal(object, output);
    }
    
    @Override
    public void marshal(T object, ContentHandler contentHandler, AttachmentMarshaller am) throws JAXBException {
        this.delegate.marshal(object, contentHandler, am);
    }
    
    @Override
    public void marshal(T object, Result result) throws JAXBException {
        this.delegate.marshal(object, result);
    }
    
    @Override
    public T unmarshal(XMLStreamReader in, AttachmentUnmarshaller au) throws JAXBException {
        return this.delegate.unmarshal(in, au);
    }
    
    @Override
    public T unmarshal(Source in, AttachmentUnmarshaller au) throws JAXBException {
        return this.delegate.unmarshal(in, au);
    }
    
    @Override
    public T unmarshal(InputStream in) throws JAXBException {
        return this.delegate.unmarshal(in);
    }
    
    @Override
    public T unmarshal(Node n, AttachmentUnmarshaller au) throws JAXBException {
        return this.delegate.unmarshal(n, au);
    }
    
    @Override
    public TypeInfo getTypeInfo() {
        return this.delegate.getTypeInfo();
    }
    
    @Override
    public boolean supportOutputStream() {
        return this.delegate.supportOutputStream();
    }
    
    public static RepeatedElementBridge.CollectionHandler create(TypeInfo ti) {
        Class<Object> javaClass = (Class)ti.type;
        if (javaClass.isArray()) {
            return new RepeatedElementBridge.ArrayHandler((Class)ti.getItemType().type);
        }
        if (List.class.equals(javaClass) || Collection.class.equals(javaClass)) {
            return RepeatedElementBridge.ListHandler;
        }
        if (Set.class.equals(javaClass) || HashSet.class.equals(javaClass)) {
            return RepeatedElementBridge.HashSetHandler;
        }
        return new BaseCollectionHandler(javaClass);
    }
    
    static {
        ListHandler = new RepeatedElementBridge.BaseCollectionHandler(List.class) {
            @Override
            public Object convert(List list) {
                return list;
            }
        };
        HashSetHandler = new RepeatedElementBridge.BaseCollectionHandler(HashSet.class) {
            @Override
            public Object convert(List list) {
                return new HashSet(list);
            }
        };
    }
    
    static class BaseCollectionHandler implements RepeatedElementBridge.CollectionHandler
    {
        Class type;
        
        BaseCollectionHandler(Class c) {
            this.type = c;
        }
        
        @Override
        public int getSize(Object c) {
            return ((Collection)c).size();
        }
        
        @Override
        public Object convert(List list) {
            try {
                Object o = this.type.newInstance();
                ((Collection)o).addAll(list);
                return o;
            }
            catch (Exception e) {
                e.printStackTrace();
                return list;
            }
        }
        
        @Override
        public Iterator iterator(Object c) {
            return ((Collection)c).iterator();
        }
    }
    
    static class ArrayHandler implements RepeatedElementBridge.CollectionHandler
    {
        Class componentClass;
        
        public ArrayHandler(Class component) {
            this.componentClass = component;
        }
        
        @Override
        public int getSize(Object c) {
            return Array.getLength(c);
        }
        
        @Override
        public Object convert(List list) {
            Object array = Array.newInstance(this.componentClass, list.size());
            for (int i = 0; i < list.size(); ++i) {
                Array.set(array, i, list.get(i));
            }
            return array;
        }
        
        @Override
        public Iterator iterator(Object c) {
            return new Iterator() {
                int index = 0;
                
                @Override
                public boolean hasNext() {
                    return c != null && Array.getLength(c) != 0 && this.index != Array.getLength(c);
                }
                
                @Override
                public Object next() throws NoSuchElementException {
                    Object retVal = null;
                    try {
                        retVal = Array.get(c, this.index++);
                    }
                    catch (ArrayIndexOutOfBoundsException ex) {
                        throw new NoSuchElementException();
                    }
                    return retVal;
                }
                
                @Override
                public void remove() {
                }
            };
        }
    }
    
    public interface CollectionHandler
    {
        public abstract int getSize(Object p0);
        
        public abstract Iterator iterator(Object p0);
        
        public abstract Object convert(List p0);
    }
}
